/*
 * Copyright (C) 2011, Jamie Thompson
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "CSVSymbianEventLogParser.h"
#include "EventTypes/PhoneCall.h"
#include "Settings.h"

#include <QDebug>

#include <QFile>
#include <QString>
#include <QStringList>

#include <stdexcept>

using namespace EventParsers;

class SortByValueDesc
{
public:
	inline bool operator()(const QPair<QChar, uint> &a, const QPair<QChar, uint> &b) const
	{
		return b.second < a.second;
	}
};

const QString ExtractString(const QString &originalString)
{
	QRegExp content("^[\"\']?(.*)?[\"\']?$");
	content.indexIn(originalString.trimmed());
	return content.cap(1);
}

iEventParser *CSVSymbianEventLogParser::IsValid(const Settings &currentSettings, QFile &eventFile)
{
	qDebug() << "Checking if a CSV call log file...";

	QTextStream stream(&eventFile);

	QString firstLineContent(stream.readLine());
	eventFile.seek(0);
	if(firstLineContent.length() > 0)
	{
		// Count the non-alphanumeric characters used
		QHash<QChar, uint> counts;
		foreach(const QChar c, firstLineContent)
			++counts[c];

		QList<QPair<QChar, uint> > orderedCounts;
		orderedCounts.reserve(counts.size());
		foreach(const QChar c, counts.keys())
			if(!QChar(c).isLetterOrNumber())
				orderedCounts.append(QPair<QChar, uint>(c, counts.value(c)));

		qSort(orderedCounts.begin(), orderedCounts.end(), SortByValueDesc());

		// Work around Q_FOREACH macro limitation when dealing with
		// multi-typed templates (comma issue)
		typedef QPair<QChar, uint> bodge;
		foreach(bodge count, orderedCounts)
			qDebug() << count.first << " = " << count.second;

		QChar delim;
		// No-one would be mad enough to use quotation marks or apostrophes
		// as their delimiter,but just in case, check the second most
		// frequent character is present thr right number of times for
		// the qutation marks to be present on every column heading (two
		// per heading, less one as they're seperators)
		if((orderedCounts.value(0).first == '"' || orderedCounts.value(0).first == '\'')
			&& ((orderedCounts.value(0).second / 2) - 1 == orderedCounts.value(1).second ))
		{
			// We're good.
			delim = orderedCounts.value(1).first;
		}
		else
			delim = orderedCounts.value(0).first;

		// Check we have the essential fields we need, and grab their
		// column ordering
		QStringList requiredHeadings;
		requiredHeadings << "etype" << "etime" << "remote" << "direction"
						 << "duration" << "number" << "data";

		EventParsers::CSVSymbianEventLogParser::ColumnIndicesHash headingPositions;
		headingPositions.reserve(requiredHeadings.count());

		QStringList headings(QString(firstLineContent).split(delim, QString::KeepEmptyParts, Qt::CaseSensitive));
		int numColumnsPerRecord(headings.count());
		for(QStringList::size_type i(0); i < headings.count(); ++i)
		{
			QString heading(ExtractString(headings.value(i)));
			qDebug() << headings.value(i) << " : " << heading;

			// Check over the required headings
			foreach(QString requiredHeading, requiredHeadings)
			{
				if(heading.toLower() == requiredHeading)
				{
					headingPositions[requiredHeading] = i;
					requiredHeadings.removeOne(requiredHeading);
				}
			}
		}

		// If we found all of the required headings, continue
		if(requiredHeadings.count() == 0)
			return new EventParsers::CSVSymbianEventLogParser(currentSettings, eventFile.fileName(), delim, numColumnsPerRecord, headingPositions);
	}

	return NULL;
}

CSVSymbianEventLogParser::CSVSymbianEventLogParser(const Settings &settings, const QString &filename, const QChar delimiter, const int numColumnsPerRecord, const ColumnIndicesHash &headingIndices)
	: m_Settings(settings), m_Delimiter(delimiter), m_NumColumnsPerRecord(numColumnsPerRecord), m_HeadingIndices(headingIndices)
{
}

EventTypes::EventFromFileList CSVSymbianEventLogParser::ParseFile(QFile &eventFile, const QList<uint> &recordsToReturn)
{
	qDebug() << "CSV Parsing NYI!";
	EventTypes::EventFromFileList fileEvents;
	//return fileEvents;

	QSet<uint> recordsToReturnSet(QSet<uint>::fromList(recordsToReturn));

	uint lineNumber(0);
	uint recordNumber(0);
	eventFile.seek(0);

	// Read the first line
	QTextStream stream(&eventFile);
	QString firstLineContent(stream.readLine());
	QStringList firstLineValues(QString(firstLineContent).split(m_Delimiter));
	if(firstLineValues.count() != m_NumColumnsPerRecord)
		throw new std::runtime_error(QString("Unexpected number of columns (%1, expected %2) on line %3 of %4")
			.arg(firstLineValues.count())
			.arg(m_NumColumnsPerRecord)
			.arg(lineNumber)
			.arg(eventFile.fileName()).toStdString());
	++lineNumber;

	// Read the main body of the file
	while(!stream.atEnd())
	{
		QStringList lineValues(QString(stream.readLine()).split(m_Delimiter));
		++lineNumber;
		// Make sure we have enough columns (i.e. handle newlines in values)
		while(lineValues.count() < m_NumColumnsPerRecord)
		{
			lineValues.append(QString(stream.readLine()).split(m_Delimiter));
			++lineNumber;
		}

		if(recordsToReturnSet.count() == 0 || recordsToReturnSet.contains(recordNumber))
		{
			bool bOK(false);
			int eType(lineValues.at(m_HeadingIndices.value("etype")).toUInt(&bOK));
			// We're only interested in phone calls
			if(bOK && eType == 0)
			{
				qDebug() << "Parsing event from line #" << lineNumber << ". Values: " << lineValues;

				QDateTime eTime(QDateTime::fromString(lineValues.at(m_HeadingIndices.value("etime")), "dd/MM/yyyy h:mm:ss ap"));
				int duration(lineValues.at(m_HeadingIndices.value("duration")).toInt(&bOK));
				if(!bOK)
				Settings::eDirection direction(lineValues.at(m_HeadingIndices.value("direction")) == "0"
					? Settings::INCOMING
					: Settings::OUTGOING);
				{
					qDebug() << QString("Unable to parse '%1' as a duration. Skipping record.")
									.arg(lineValues.at(m_HeadingIndices.value("duration")));
					continue;
				}
				QString number(ExtractString(lineValues.at(m_HeadingIndices.value("number"))));
				QString data(ExtractString(lineValues.at(m_HeadingIndices.value("data"))));

				QSharedPointer<EventTypes::iEvent> newEvent(new EventTypes::PhoneCall(
					CurrentSettings(),
					direction,
					eTime,
					number,
					duration));
				fileEvents.append(EventTypes::EventFromFile(newEvent, recordNumber));
			}
		}

		++recordNumber;
	}

	qDebug() << QString("File pos: %1, bAvail: %2, canReadLine: %3").arg(eventFile.pos()).arg(eventFile.bytesAvailable()).arg(eventFile.canReadLine());
	qDebug() << fileEvents.count() << " events loaded from file";
	return fileEvents;
}
